from typing import Any

from pydantic import AnyHttpUrl, BaseSettings, PostgresDsn, validator


def convert_env_to_list(cls, v: str | list[str]) -> list[str] | str:
    if isinstance(v, str) and not v.startswith("["):
        return [i.strip() for i in v.split(",")]
    elif isinstance(v, list):
        return v
    raise ValueError(v)


class ServerSettings(BaseSettings):
    API_STR: str = "/api"
    SERVER_NAME: str
    SERVER_HOST: AnyHttpUrl
    BACKEND_CORS_ORIGINS: list[AnyHttpUrl] = [
        "http://localhost:4200",
        "http://0.0.0.0:4200",
    ]

    # validators
    _convert_to_list = validator("BACKEND_CORS_ORIGINS", allow_reuse=True)(convert_env_to_list)

    class Config:
        case_sensitive = True


class AzureSettings(BaseSettings):
    AAD_APP_CLIENT_ID: str
    AAD_TENANT_ID: str
    AAD_TEST_CLIENT_SECRET: str | None

    class Config:
        case_sensitive = True


class PostgresSettings(BaseSettings):
    POSTGRES_HOST: str
    POSTGRES_USER: str
    POSTGRES_PASSWORD: str
    POSTGRES_DB: str
    POSTGRES_PORT: str
    POSTGRES_SSL: bool = False
    POSTGRES_MAX_OVERFLOW = 30
    POSTGRES_POOL_SIZE: int = 20
    SQLALCHEMY_DATABASE_URI: PostgresDsn | None = None

    @validator("SQLALCHEMY_DATABASE_URI", pre=True)
    def assemble_db_connection(cls, v: str | None, values: dict[str, Any]) -> Any:
        if isinstance(v, str):
            return v
        return PostgresDsn.build(
            scheme="postgresql+asyncpg",
            user=values.get("POSTGRES_USER"),
            password=values.get("POSTGRES_PASSWORD"),
            host=values.get("POSTGRES_HOST"),
            path=f"/{values.get('POSTGRES_DB') or ''}",
            port=values.get("POSTGRES_PORT"),
        )

    class Config:
        case_sensitive = True


class EmailSettings(BaseSettings):
    SENDGRID_SECRET: str
    EMAIL_NOTIFICATION_FROM: str
    INTERNAL_EMAIL_DOMAINS_LIST: str | None = None
    DEFAULT_AD_FQDN: str

    # validators
    _convert_to_list = validator("INTERNAL_EMAIL_DOMAINS_LIST", allow_reuse=True)(convert_env_to_list)

    class Config:
        case_sensitive = True


class MetricsSettings(BaseSettings):
    OTLP_GRPC_ENDPOINT: AnyHttpUrl = "http://tempo-distributor.monitoring:4317"
    ENABLE_METRICS: bool = True
    ENABLE_TELEMETRY: bool = True

    class Config:
        case_sensitive = True


class Settings(ServerSettings, AzureSettings, PostgresSettings, EmailSettings, MetricsSettings):
    ROUTER_URL: AnyHttpUrl
    AAD_GRAPH_SECRET: str

    class Config:
        case_sensitive = True



from fastapi import Depends, Security

from arkitema_config.security import azure_scheme

try:
    from core.config import settings
except (ImportError, ModuleNotFoundError):
    from arkitema_config import config

    settings = config.Settings()


try:
    from arkitema_config.connection import get_db

    db_url = settings.SQLALCHEMY_DATABASE_URI

    async def get_context(session=Depends(get_db), user=Security(azure_scheme)):
        return {"session": session, "user": user}

except (ImportError, ModuleNotFoundError, AttributeError):

    async def get_context(user=Security(azure_scheme)):
        return {"user": user}



from fastapi_azure_auth import SingleTenantAzureAuthorizationCodeBearer

try:
    from core.config import settings
except (ImportError, ModuleNotFoundError):
    from arkitema_config import config

    settings = config.Settings()

azure_scheme = SingleTenantAzureAuthorizationCodeBearer(
    app_client_id=settings.AAD_APP_CLIENT_ID,
    tenant_id=settings.AAD_TENANT_ID,
    auto_error=False,
    allow_guest_users=True,
    scopes={
        f"api://{settings.AAD_APP_CLIENT_ID}/user_impersonation": "user_impersonation",
    },
)



from typing import Generic, Optional, TypeVar

import strawberry

GenericType = TypeVar("GenericType")


@strawberry.type
class Connection(Generic[GenericType]):
    """Represents a paginated relationship between two entities

    This pattern is used when the relationship itself has attributes.
    In a Facebook-based domain example, a friendship between two people
    would be a connection that might have a `friendshipStartTime`
    """

    page_info: "PageInfo"
    edges: list["Edge[GenericType]"]
    num_edges: int


@strawberry.type
class PageInfo:
    """Pagination context to navigate objects with cursor-based pagination

    Instead of classic offset pagination via `page` and `limit` parameters,
    here we have a cursor of the last object, and we fetch items starting from that one

    Read more at:
        - https://graphql.org/learn/pagination/#pagination-and-edges
        - https://relay.dev/graphql/connections.htm
    """

    has_next_page: bool
    has_previous_page: bool
    start_cursor: Optional[str]
    end_cursor: Optional[str]


@strawberry.type
class Edge(Generic[GenericType]):
    """An edge may contain additional information of the relationship. This is the trivial case"""

    node: GenericType
    cursor: str


Cursor = str



import asyncio
import logging
import logging.config
import os
from pathlib import Path

import asyncpg
from sqlmodel.ext.asyncio.session import AsyncSession

from arkitema_config.connection import create_postgres_engine

try:
    from core.config import settings
except (ImportError, ModuleNotFoundError):
    from arkitema_config import config

    settings = config.Settings()

logging.config.fileConfig(Path(__file__).parent.parent / "logging.conf", disable_existing_loggers=False)
logger = logging.getLogger(__name__)


async def main():
    sql_folder = Path("/app/db_backups")
    sql_folder.mkdir(exist_ok=True)
    logger.info(f"Created folder {sql_folder}")

    sql_file = sql_folder / f"{settings.POSTGRES_DB}.sql"
    db_config = {
        "user": os.getenv("AZURE_POSTGRES_USER"),
        "host": os.getenv("AZURE_POSTGRES_HOST"),
        "port": os.getenv("AZURE_POSTGRES_PORT"),
        "database": settings.POSTGRES_DB,
        "password": os.getenv("AZURE_POSTGRES_PASSWORD"),
    }
    await download_latest_db(sql_file, db_config)
    await prepare_db(sql_file)
    await load_latest_db(sql_file)


async def download_latest_db(sql_path: Path, db_config: dict):
    cmd = (
        f"pg_dump -h {db_config.get('host')} -U {db_config.get('user')} "
        f"-p {db_config.get('port')} -d {db_config.get('database')} > {str(sql_path)}"
    )

    logger.info("Starting downloading database")

    process = await asyncio.create_subprocess_shell(
        cmd,
        stderr=asyncio.subprocess.PIPE,
        stdout=asyncio.subprocess.PIPE,
        env={"PGPASSWORD": db_config.get("password")},
    )

    stdout, stderr = await process.communicate()

    logger.info("Finished downloading")
    if stdout:
        logger.info(stdout.decode())
    if stderr:
        logger.error(stderr.decode())


async def load_latest_db(sql_path: Path):
    cmd = (
        f"psql -h {settings.POSTGRES_HOST} -U {settings.POSTGRES_USER} "
        f"-p {settings.POSTGRES_PORT} -d {settings.POSTGRES_DB} -f {str(sql_path)}"
    )

    logger.info("Starting loading database")

    process = await asyncio.create_subprocess_shell(
        cmd,
        stderr=asyncio.subprocess.PIPE,
        stdout=asyncio.subprocess.PIPE,
        env={"PGPASSWORD": settings.POSTGRES_PASSWORD},
    )

    stdout, stderr = await process.communicate()

    logger.info("Loaded database")
    if stdout:
        logger.info(stdout.decode())
    if stderr:
        logger.error(stderr.decode())


async def create_azure_roles(connection):
    logger.info("Creating Azure roles")
    await connection.execute("CREATE ROLE azure_pg_admin")
    await connection.execute("CREATE ROLE lcadbadmin")


async def prepare_db(sql_file: Path):
    logger.info(f"Loading downloaded database: {sql_file}")
    conn = await asyncpg.connect(
        user=settings.POSTGRES_USER,
        database="postgres",
        port=settings.POSTGRES_PORT,
        host=settings.POSTGRES_HOST,
        password=settings.POSTGRES_PASSWORD,
    )
    try:
        logger.info("Established connection to database")
        await conn.execute(
            f"SELECT pg_terminate_backend(pg_stat_activity.pid) FROM pg_stat_activity WHERE pg_stat_activity.datname = '{settings.POSTGRES_DB}' AND pid <> pg_backend_pid()"
        )
        logger.info("Dropping database")
        await conn.execute(f"DROP DATABASE IF EXISTS {settings.POSTGRES_DB}")
        logger.info("Creating new database")
        await conn.execute(f"CREATE DATABASE {settings.POSTGRES_DB}")
        logger.info("Closed connection to database")
        await create_azure_roles(conn)

    except Exception as e:
        logger.error(e)
        raise e
    finally:
        await conn.close()



import json
import re
from datetime import datetime

from aiocache import Cache
from msgraph.core import GraphClient
from requests import Response

from arkitema_config import exceptions

try:
    from core.config import settings
except (ImportError, ModuleNotFoundError):
    from arkitema_config import config

    settings = config.Settings()

cache = Cache()


async def get_aad_user_by_email(email: str) -> dict[str, str]:
    """Check if user exists in Azure Active Directory"""
    user_data = await cache.get(email, namespace="azure_emails")
    if user_data:
        return user_data

    graph = GraphClient(credential=settings.AAD_GRAPH_SECRET)
    headers = {"Content-Type": "application/json"}

    # construct user principal name
    user_principal_name = email
    # check if user is an external user
    if not any(domain in email for domain in settings.INTERNAL_EMAIL_DOMAINS_LIST):
        # externals have principal name formatted like
        # xxxx_gmail.com%23EXT%23@cowi.onmicrosoft.com
        user_principal_name = user_principal_name.replace("@", "_")
        user_principal_name += f"%23EXT%23@{settings.DEFAULT_AD_FQDN}"
    response = graph.get(url=f"/users/{user_principal_name}", headers=headers)

    data = response.json()
    await cache.add(email, data, namespace="azure_emails", ttl=60 * 5)
    return data


def invite_user_to_aad(email: str, name: str, platform_url: str) -> Response:
    """
    invites a user to organization's Active Directory

    Parameters
    ----------
    email: str
        user's email to create an AD account and send invitation to
    name: str
        name to be displayed on AD and on the platform
    platform_url: str
        platform url that the invitation will redirect to

    Returns
    -------
    response: requests.Response
        http response received from Graph API
    """

    graph = GraphClient(credential=settings.AAD_GRAPH_SECRET)

    headers = {"Content-Type": "application/json"}

    body = {
        "invitedUserEmailAddress": email,
        "inviteRedirectUrl": platform_url,
        "invitedUserDisplayName": name,
        "sendInvitationMessage": True,
        "userPrincipalName": email,
    }
    # add user
    response = graph.post(url=f"/invitations", data=json.dumps(body), headers=headers)
    return response


async def get_users_from_azure(user_ids: str | list[str]) -> list[dict[str, str]]:
    """Fetch Users from Azure Active Directory"""

    if not user_ids:
        return [{}]

    if not isinstance(user_ids, list):
        user_ids = [user_ids]

    user_data = await cache.multi_get(user_ids, namespace="azure_users")
    users = {user_id: user_data[index] for index, user_id in enumerate(user_ids)}
    missing_users = [user_id for user_id, user_data in users.items() if user_data is None]

    requests = []
    headers = {"Content-Type": "application/json"}
    for user_id in missing_users:
        # graph api Beta supports signInActivity
        request = {
            "id": f"{user_id}",
            "method": "GET",
            "url": f"/users/{user_id}?$select=id,displayName,mail," "userPrincipalName,companyName,signInActivity",
            # signInActivity field requires AuditLog.Read.All permission
            "headers": headers,
        }

        requests.append(request)
    if requests:
        graph = GraphClient(credential=settings.AAD_GRAPH_SECRET)
        # use /beta/$batch url if fetching signInActivity
        responses = graph.post(
            url="https://graph.microsoft.com/beta/$batch",
            data=json.dumps({"requests": requests}),
            headers=headers,
        )
        if not responses.status_code == 200:
            raise exceptions.MSGraphException(f"Failed to fetch users via Graph API: {responses.text}")

        data: dict[str, list[dict]] = responses.json()

        for response in data.get("responses"):
            if not response.get("status") == 200:
                raise exceptions.MSGraphException(f"Failed to fetch the response from responses: {response}")
            body: dict[str, str] = response.get("body")
            email = body.get("mail")
            # some accounts have null emails
            if not email:
                if re.match(
                    r"^([a-zA-Z0-9_\-\.]+)@([a-zA-Z0-9_\-\.]+)\.([a-zA-Z]{2,5})$",
                    body.get("userPrincipalName", ""),
                ):
                    email = body.get("userPrincipalName")
                else:
                    email = "NA"
            if last_login := body.get("signInActivity", {}).get("lastSignInDateTime"):
                last_login = datetime.strptime(last_login, r"%Y-%m-%dT%H:%M:%SZ").date()
            users[body.get("id")] = {
                "user_id": body.get("id"),
                "name": body.get("displayName"),
                "email": email,
                "company": body.get("companyName"),
                "last_login": last_login,
            }
        await cache.multi_set(pairs=users.items(), ttl=60 * 5, namespace="azure_users")
    return list(users.values())
